<?php

namespace TC\PaygateHeleket\Payment;

use GuzzleHttp\Exception\GuzzleException;
use XF;
use XF\Entity\PaymentProfile;
use XF\Entity\PurchaseRequest;
use XF\Http\Request;
use XF\Mvc\Controller;
use XF\Payment\AbstractProvider;
use XF\Payment\CallbackState;
use XF\Purchasable\Purchase;

class Heleket extends AbstractProvider
{
	/**
	 * @return string
	 */
	public function getTitle(): string
	{
		return 'Heleket';
	}

	/**
	 * @return string
	 */
	public function getApiEndpoint(): string
	{
		return 'https://api.heleket.com/';
	}

	/**
	 * @var string
	 */
	private string $cryptoCurrency;

	/**
	 * @var string
	 */
	private string $cryptoNetwork;

	/**
	 * @param array $options
	 * @param array $errors
	 *
	 * @return bool
	 * @throws GuzzleException
	 */
	public function verifyConfig(array &$options, &$errors = []): bool
	{
		if (empty($options['merchant_id']) || empty($options['api_key']))
		{
			$errors[] = XF::phrase('tc_pg_heleket_err_you_must_provide_all_data');
		}

		$currenciesNetworks = [];

		$availableCurrencies = $this->apiCall($options['merchant_id'], $options['api_key'], 'v1/settings/payments/currencies/list', [], $error);
		if (!$error)
		{
			foreach ($availableCurrencies as $availableCurrency)
			{
				if ($availableCurrency['is_enabled'] === true)
				{
					$networkCode = strtoupper($availableCurrency['network_code']);

					$currenciesNetworks[$availableCurrency['currency_code']][$networkCode] = $networkCode;
				}
			}

			$options['currenciesNetworks'] = $currenciesNetworks;
		}
		else
		{
			$errors[] = $error;
		}

		return !$errors;
	}

	/**
	 * @param PurchaseRequest $purchaseRequest
	 * @param Purchase        $purchase
	 *
	 * @return array
	 */
	protected function getPaymentParams(PurchaseRequest $purchaseRequest, Purchase $purchase): array
	{
		$params = [
			'amount'          => number_format($purchaseRequest->cost_amount, 2, '.', ''),
			'currency'        => $purchaseRequest->cost_currency,
			'order_id'        => $purchaseRequest->request_key,
			'additional_data' => json_encode([
				'url_return' => $purchase->returnUrl
			]),
			'url_return'      => $purchase->returnUrl,
			'url_callback'    => $this->getCallbackUrl()
		];

		if (!empty($this->cryptoCurrency) && !empty($this->cryptoNetwork))
		{
			$params += [
				'to_currency' => $this->cryptoCurrency,
				'network'     => $this->cryptoNetwork,
			];
		}

		return $params;
	}

	/**
	 * @param Controller      $controller
	 * @param PurchaseRequest $purchaseRequest
	 * @param Purchase        $purchase
	 *
	 * @return XF\Mvc\Reply\Error|XF\Mvc\Reply\Redirect
	 */
	public function initiatePayment(Controller $controller, PurchaseRequest $purchaseRequest, Purchase $purchase): XF\Mvc\Reply\AbstractReply
	{
		$profileOptions = $purchaseRequest->PaymentProfile->options;

		if ($profileOptions['type'] == 'redirect' && !empty($profileOptions['disable_currency_selector']))
		{
			return $this->processPayment($controller, $purchaseRequest, $purchaseRequest->PaymentProfile, $purchase);
		}

		$viewParams = [
			'currenciesNetworks' => $profileOptions['currenciesNetworks'],
			'purchaseRequest'    => $purchaseRequest
		];

		return $controller->view('TC\PaygateHeleket:Payment\Initiate', 'tc_payment_initiate_heleket', $viewParams);
	}

	/**
	 * @param Controller      $controller
	 * @param PurchaseRequest $purchaseRequest
	 * @param PaymentProfile  $paymentProfile
	 * @param Purchase        $purchase
	 *
	 * @return XF\Mvc\Reply\AbstractReply
	 */
	public function processPayment(Controller $controller, PurchaseRequest $purchaseRequest, PaymentProfile $paymentProfile, Purchase $purchase): XF\Mvc\Reply\AbstractReply
	{
		$profileOptions = $paymentProfile->options;
		$request = $controller->request();

		// Payment UUID doesn't exist -> creating payment in Heleket
		if (empty($purchaseRequest->provider_metadata))
		{
			$this->cryptoCurrency = $request->filter('currency', 'str');
			$this->cryptoNetwork = $request->filter('network', 'str');

			if (
				($profileOptions['type'] == 'redirect' && !empty($profileOptions['disable_currency_selector']))
				|| (!empty($this->cryptoCurrency) && !empty($this->cryptoNetwork))
			)
			{
				$payment = $this->getPaymentParams($purchaseRequest, $purchase);

				$invoice = $this->apiCall($profileOptions['merchant_id'], $profileOptions['api_key'], 'v1/payment', $payment, $error);
				if ($error)
				{
					return $controller->error($error);
				}

				// Save payment UUID and redirect user
				$purchaseRequest->fastUpdate('provider_metadata', $invoice['uuid']);

				$purchaseRequest->fastUpdate('extra_data', array_merge($purchaseRequest->extra_data, [
					'crypto_amount'   => $invoice['payer_amount'],
					'crypto_currency' => $invoice['payer_currency'],
					'crypto_network'  => $invoice['network']
				]));

				$returnType = $profileOptions['type'] ?? 'internal';

				if ($returnType == 'redirect')
				{
					return $controller->redirect($invoice['url']);
				}

				if (!empty($invoice['address']))
				{
					return $controller->rerouteController(
						\TC\PaygateHeleket\Pub\Controller\Heleket::class,
						'status',
						[
							'request_key' => $purchaseRequest->request_key
						]
					);
				}

				return $controller->error(XF::phrase('something_went_wrong_please_try_again'));
			}

			return $controller->error(XF::phrase('tc_pg_heleket_errors_invalid_currency_or_network'));
		}

		return $controller->redirect($controller->buildLink('tc-pg-heleket/status', ['request_key' => $purchaseRequest->request_key]));
	}

	/**
	 * @param Request $request
	 *
	 * @return CallbackState
	 */
	public function setupCallback(Request $request): CallbackState
	{
		$state = new CallbackState();

		$state->providerId = $request->filter('_xfProvider', 'str');
		$state->ip = $request->getIp();
		$state->rawInput = $request->getInputRaw();

		$state->inputArray = json_decode($state->rawInput, true) ?? [];
		$state->inputFiltered = $request->getInputFilterer()->filterArray($state->inputArray, [
			'type'                => 'str',
			'uuid'                => 'str',
			'order_id'            => 'str',
			'amount'              => 'str',
			'payment_amount'      => 'str',
			'payment_amount_usd'  => 'str',
			'merchant_amount'     => 'str',
			'commission'          => 'str',
			'is_final'            => 'bool',
			'status'              => 'str',
			'from'                => 'str',
			'wallet_address_uuid' => 'str',
			'network'             => 'str',
			'currency'            => 'str',
			'payer_currency'      => 'str',
			'additional_data'     => 'str',
			'convert'             => 'str',
			'sign'                => 'str'
		]);

		$state->transactionId = $state->inputFiltered['uuid'] ?? null;
		$state->requestKey = $state->inputFiltered['order_id'] ?? null;

		$state->signature = $state->inputFiltered['sign'] ?? '';

		$state->costAmount = $state->inputFiltered['amount'] ?? null;
		$state->costCurrency = $state->inputFiltered['currency'] ?? null;

		$state->status = $state->inputFiltered['status'] ?? null;
		$state->isFinal = $state->inputFiltered['is_final'] ?? null;

		return $state;
	}

	/**
	 * @param CallbackState $state
	 *
	 * @return bool
	 */
	public function validateCallback(CallbackState $state): bool
	{
		// Additional check because XF doesn't do it. Why, Chris?
		if ($state->transactionId && $state->requestKey)
		{
			if ($state->getPaymentProfile()->provider_id == $state->providerId)
			{
				return parent::validateCallback($state);
			}

			$state->logType = 'info';
			$state->logMessage = 'Invalid provider.';

			return false;
		}

		$state->logType = 'info';
		$state->logMessage = 'No Transaction ID or Request Key. No action to take.';

		return false;
	}

	/**
	 * @param CallbackState $state
	 *
	 * @return bool
	 */
	public function validatePurchasableData(CallbackState $state): bool
	{
		$paymentProfile = $state->getPaymentProfile();
		$profileOptions = $paymentProfile->options;

		$options = $paymentProfile->options;
		if (!empty($options['api_key']))
		{
			if (!empty($state->signature))
			{
				$signatureArray = $state->inputArray;
				unset($signatureArray['sign']);

				$generatedSignature = $this->getSignature(json_encode($signatureArray, JSON_UNESCAPED_UNICODE), $profileOptions['api_key']);
				if (hash_equals($state->signature, $generatedSignature))
				{
					return true;
				}

				$state->logType = 'error';
				$state->logMessage = 'Invalid signature.';

				return false;
			}

			$state->logType = 'error';
			$state->logMessage = 'Empty signature.';

			return false;
		}

		$state->logType = 'error';
		$state->logMessage = 'Invalid api_key in forum settings.';

		return false;
	}

	/**
	 * @param CallbackState $state
	 */
	public function getPaymentResult(CallbackState $state): void
	{
		if ($state->isFinal
			&& (strtolower($state->status) == 'paid' || strtolower($state->status) == 'paid_over'))
		{
			$state->paymentResult = CallbackState::PAYMENT_RECEIVED;
		}
	}

	/**
	 * @param CallbackState $state
	 */
	public function prepareLogData(CallbackState $state): void
	{
		$state->logDetails = [
			'ip'            => $state->ip,
			'request_time'  => XF::$time,
			'inputFiltered' => $state->inputFiltered,
			'inputJson'     => $state->rawInput
		];
	}

	/**
	 * @var array
	 */
	protected array $supportedCurrencies = [
		'USD', 'EUR', 'GBP', 'RUB', 'UAH', 'BYN'
	];

	/**
	 * @param PaymentProfile $paymentProfile
	 * @param                $unit
	 * @param                $amount
	 * @param int            $result
	 *
	 * @return bool
	 */
	public function supportsRecurring(PaymentProfile $paymentProfile, $unit, $amount, &$result = self::ERR_NO_RECURRING): bool
	{
		return false;
	}

	/**
	 * @return array
	 */
	protected function getSupportedRecurrenceRanges(): array
	{
		return [];
	}

	/**
	 * @param PaymentProfile $paymentProfile
	 * @param                $currencyCode
	 *
	 * @return bool
	 */
	public function verifyCurrency(PaymentProfile $paymentProfile, $currencyCode): bool
	{
		$addOns = XF::app()->container('addon.cache');

		if (!empty($paymentProfile->options['tc_cu_exchange_to'])
			&& array_key_exists('TC/CurrencyUtils', $addOns)
			&& XF::registry()->exists('tcCurrencies'))
		{
			return array_key_exists($currencyCode, XF::registry()->get('tcCurrencies'));
		}

		return in_array($currencyCode, $this->supportedCurrencies);
	}

	private function getRefCode(): string
	{
		return 'wLKK97';
	}

	/**
	 * @param string $merchantId
	 * @param string $apiKey
	 * @param string $path
	 * @param array  $data
	 * @param        $error
	 *
	 * @return array
	 * @throws GuzzleException
	 */
	public function apiCall(string $merchantId, string $apiKey, string $path, array $data, &$error = null): array
	{
		$dataString = json_encode($data, JSON_UNESCAPED_UNICODE);

		$response = XF::app()->http()->client()->post($this->getApiEndpoint() . $path, [
			'headers'    => [
				'merchant'     => $merchantId,
				'sign'         => $this->getSignature($dataString, $apiKey),
				'Content-Type' => 'application/json'
			],
			'body'       => $dataString,
			'exceptions' => false
		]);

		if ($response)
		{
			$responseData = json_decode($response->getBody(), true);

			$state = $responseData['state'] ?? 0;
			if ($response->getStatusCode() == 200 && is_array($responseData) && $state == 0)
			{
				return $responseData['result'] ?? [];
			}
			else
			{
				if (!empty($responseData['message']))
				{
					$error = $responseData['message'];

					return [];
				}

				if (!empty($responseData['errors']))
				{
					$error = is_array($responseData['errors']) ? json_encode($responseData['errors']) : $responseData['errors'];

					return [];
				}

				$error = XF::phrase('something_went_wrong_please_try_again')->render();

				return [];
			}
		}

		$error = XF::phrase('something_went_wrong_please_try_again')->render();

		return [];
	}

	/**
	 * @param string $data
	 * @param string $apiKey
	 *
	 * @return string
	 */
	public function getSignature(string $data, string $apiKey): string
	{
		return hash('md5', base64_encode($data) . $apiKey);
	}

	/**
	 * @param $network
	 * @param $address
	 * @param $amount
	 *
	 * @return string
	 */
	public function generatePaymentLink($network, $address, $amount): string
	{
		if ($network == 'ton')
		{
			$amount = number_format($amount, 9, '', '');

			return 'https://app.tonkeeper.com/transfer/' . $address . '?amount=' . $amount;
		}

		return $network . ':' . $address . '?amount=' . $amount;
	}
}
